"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PCSourceWriter = void 0;
var mutations_1 = require("./mutations");
var paperclip_utils_1 = require("paperclip-utils");
var string_editor_1 = require("./string-editor");
var ANNOTATION_KEYS = ["title", "width", "height", "x", "y"];
var PCSourceWriter = /** @class */ (function () {
    function PCSourceWriter(_engine) {
        this._engine = _engine;
    }
    PCSourceWriter.prototype.apply = function (mutations) {
        var changes = this._getContentChanges(mutations);
        for (var uri in changes) {
            var newContent = string_editor_1.editString(this._engine.getVirtualContent(uri), changes[uri]);
            this._engine.updateVirtualFileContent(uri, newContent);
        }
        // return changes so that
        return changes;
    };
    PCSourceWriter.prototype._getContentChanges = function (mutations) {
        var _a;
        var changes = [];
        var engine = this._engine;
        for (var _i = 0, mutations_2 = mutations; _i < mutations_2.length; _i++) {
            var _b = mutations_2[_i], targetId = _b.targetId, action = _b.action;
            var _c = engine.getExpressionById(targetId), uri = _c[0], targetAst = _c[1];
            var documentAst = engine.getLoadedAst(uri);
            var textSource = { uri: uri, location: targetAst.location };
            switch (action.kind) {
                case mutations_1.PCMutationActionKind.ANNOTATIONS_CHANGED: {
                    changes.push(this._getAnnotationChange(textSource, (_a = paperclip_utils_1.getPCNodeAnnotations(paperclip_utils_1.getNodeById(targetAst.id, documentAst), documentAst)) === null || _a === void 0 ? void 0 : _a.location, action.annotations));
                    break;
                }
                case mutations_1.PCMutationActionKind.CSS_DECLARATION_CHANGED: {
                    changes.push(this._getCSSDeclarationChange(textSource, action));
                    break;
                }
                case mutations_1.PCMutationActionKind.EXPRESSION_DELETED: {
                    changes.push.apply(changes, this._getExpressionDeletedChanged(textSource, targetAst));
                    break;
                }
            }
        }
        var changesByUri = {};
        for (var _d = 0, changes_1 = changes; _d < changes_1.length; _d++) {
            var change = changes_1[_d];
            if (!changesByUri[change.uri]) {
                changesByUri[change.uri] = [];
            }
            changesByUri[change.uri].push(change);
        }
        return changesByUri;
    };
    PCSourceWriter.prototype._getExpressionDeletedChanged = function (exprTextSource, ast) {
        var node = getAssocNode(exprTextSource, ast);
        var parent = paperclip_utils_1.getParentNode(node, ast);
        var childIndex = parent.children.findIndex(function (child) { return child === node; });
        var changes = [];
        var beforeChild = childIndex > 0 ? parent.children[childIndex - 1] : null;
        // if before child is a comment, then assume it's an annotation
        if (beforeChild && beforeChild.nodeKind === paperclip_utils_1.NodeKind.Comment) {
            changes.push({
                uri: exprTextSource.uri,
                start: beforeChild.location.start,
                end: beforeChild.location.end,
                value: ""
            });
        }
        changes.push({
            uri: exprTextSource.uri,
            start: exprTextSource.location.start,
            end: exprTextSource.location.end,
            value: ""
        });
        return changes;
    };
    PCSourceWriter.prototype._getCSSDeclarationChange = function (exprTextSource, action) {
        return {
            uri: exprTextSource.uri,
            start: exprTextSource.location.start,
            end: exprTextSource.location.end,
            value: action.name + ": " + action.value + ";"
        };
    };
    PCSourceWriter.prototype._getAnnotationChange = function (exprTextSource, annottaionsLocation, annotations) {
        var buffer = ["<!--\n"];
        for (var key in annotations) {
            var chunk = ["  @" + key + " "];
            var value = annotations[key];
            if (Array.isArray(value) ||
                typeof value === "string" ||
                typeof value === "number") {
                chunk.push(JSON.stringify(value));
            }
            else {
                chunk.push("{ ");
                var items = [];
                var sortedKeys = Object.keys(value).sort(function (a, b) {
                    return ANNOTATION_KEYS.indexOf(a) < ANNOTATION_KEYS.indexOf(b)
                        ? -1
                        : 1;
                });
                for (var _i = 0, sortedKeys_1 = sortedKeys; _i < sortedKeys_1.length; _i++) {
                    var k = sortedKeys_1[_i];
                    items.push(k + ": " + JSON.stringify(value[k]));
                }
                chunk.push(items.join(", "), " }");
            }
            buffer.push(chunk.join(""), "\n");
        }
        buffer.push("-->");
        // insertion - give it some padding
        if (!annottaionsLocation) {
            buffer.unshift("\n");
            buffer.push("\n");
        }
        return {
            uri: exprTextSource.uri,
            start: annottaionsLocation
                ? annottaionsLocation.start
                : exprTextSource.location.start,
            end: annottaionsLocation
                ? annottaionsLocation.end
                : exprTextSource.location.start,
            // newline may have been clipped off, so re-add if that happens
            value: buffer.join("")
        };
    };
    return PCSourceWriter;
}());
exports.PCSourceWriter = PCSourceWriter;
var getAssocNode = function (exprTextSource, root) {
    var foundExpr;
    paperclip_utils_1.traverseExpression(root, function (node) {
        if (node.location.start === exprTextSource.location.start &&
            node.location.end === exprTextSource.location.end) {
            foundExpr = node;
            return false;
        }
    });
    // should NOT happen
    if (!foundExpr) {
        console.error("[PCSourceWriter] Cannot find associated node, content is likely out of sync with visual editor.");
    }
    return foundExpr;
};
